@@ -1,3 +1,4 @@
+import ast
 import re
 import types
 
@@ -10,6 +11,10 @@ def Grammar(description, name='grammar', include_source=False):
     # Parse the grammar description.
     raw = meta.parse(description)
 
+    # If the grammar is just an expression, create an implicit 'start' rule.
+    if not isinstance(raw, list):
+        raw = [meta.RuleDef(is_ignored=False, name='start', params=None, expr=raw)]
+
     # Create the docstring for the module.
     docstring = '# Grammar definition:\n' + description
 
@@ -31,13 +36,34 @@ def Grammar(description, name='grammar', include_source=False):
 
 def _create_parsing_expression(node):
     if isinstance(node, meta.StringLiteral):
-        if node.ignore_case:
-            return RegexLiteral(re.escape(node.value), ignore_case=True)
+        ignore_case = node.value.endswith(('i', 'I'))
+        value = ast.literal_eval(node.value[:-1] if ignore_case else node.value)
+        if ignore_case:
+            return RegexLiteral(re.escape(value), ignore_case=True)
         else:
-            return StringLiteral(node.value)
+            return StringLiteral(value)
 
     if isinstance(node, meta.RegexLiteral):
-        return RegexLiteral(node.value, ignore_case=node.ignore_case)
+        is_binary = node.value.startswith('b')
+        ignore_case = node.value.endswith(('i', 'I'))
+        value = node.value
+
+        # Remove leading 'b'.
+        if is_binary:
+            value = value[1:]
+
+        # Remove trailing 'i'.
+        if ignore_case:
+            value = value[:-1]
+
+        # Remove backslashes.
+        value = value[1:-1]
+
+        # Enocde binary string.
+        if is_binary:
+            value = value.encode('ascii')
+
+        return RegexLiteral(value, ignore_case=ignore_case)
 
     if isinstance(node, meta.PythonExpression):
         return PythonExpression(node.value)
@@ -58,11 +84,14 @@ def _create_parsing_expression(node):
         return node
 
     if isinstance(node, meta.Postfix) and isinstance(node.operator, meta.ArgList):
-        left = node.left
+        left, args = node.left, node.operator.args
         if isinstance(left, Ref) and hasattr(parsing_expressions, left.name):
-            return getattr(parsing_expressions, left.name)(*node.operator.args)
+            return getattr(parsing_expressions, left.name)(
+                *[unwrap(x) for x in args if not isinstance(x, KeywordArg)],
+                **{x.name: unwrap(x.expr) for x in args if isinstance(x, KeywordArg)},
+            )
         else:
-            return Call(left, node.operator.args)
+            return Call(left, args)
 
     if isinstance(node, meta.Postfix):
         classes = {
@@ -73,6 +102,14 @@ def _create_parsing_expression(node):
         if isinstance(node.operator, str) and node.operator in classes:
             return classes[node.operator](node.left)
 
+        if isinstance(node.operator, meta.Repeat):
+            start = uncook(node.operator.start)
+            stop = uncook(node.operator.stop)
+            return List(node.left, min_len=start, max_len=stop)
+
+    if isinstance(node, meta.Repeat):
+        return node
+
     if isinstance(node, meta.Infix) and node.operator == '|':
         left, right = node.left, node.right
         left = list(left.exprs) if isinstance(left, Choice) else [left]
@@ -83,8 +120,8 @@ def _create_parsing_expression(node):
         classes = {
             '|>': lambda a, b: Apply(a, b, apply_left=False),
             '<|': lambda a, b: Apply(a, b, apply_left=True),
-            '/': lambda a, b: Alt(a, b, allow_trailer=True),
-            '//': lambda a, b: Alt(a, b, allow_trailer=False),
+            '/?': lambda a, b: Sep(a, b, allow_trailer=True),
+            '//': lambda a, b: Sep(a, b, allow_trailer=False),
             '<<': Left,
             '>>': Right,
             'where': Where,
@@ -100,5 +137,25 @@ def _create_parsing_expression(node):
     if isinstance(node, meta.ClassDef):
         return Class(node.name, node.params, node.fields)
 
+    if isinstance(node, meta.IgnoreStmt):
+        return Rule(None, None, node.expr, is_ignored=True)
+
     # Otherwise, fail if we don't know what to do with this node.
     raise Exception(f'Unexpected expression: {node!r}')
+
+
+def unwrap(x):
+    return eval(x.source_code) if isinstance(x, PythonExpression) else x
+
+
+def uncook(x):
+    if x is None:
+        return None
+    if isinstance(x, PythonExpression) and x.source_code == 'None':
+        return None
+    if isinstance(x, PythonExpression):
+        return x.source_code
+    if isinstance(x, Ref):
+        return x.name
+
+    raise Exception(f'Expected name or Python expression. Received: {x}')
