@@ -98,17 +98,25 @@ def visit(previsit, expr, postvisit=None):
             visit(previsit, child, postvisit)
 
 
-class Alt(Expr):
+class Sep(Expr):
     num_blocks = 2
 
-    def __init__(self, expr, separator, allow_trailer=False, allow_empty=True):
+    def __init__(
+            self,
+            expr,
+            separator,
+            discard_separators=True,
+            allow_trailer=False,
+            allow_empty=True,
+        ):
         self.expr = expr
         self.separator = separator
+        self.discard_separators = discard_separators
         self.allow_trailer = allow_trailer
         self.allow_empty = allow_empty
 
     def __str__(self):
-        op = '/' if self.allow_trailer else '//'
+        op = '/?' if self.allow_trailer else '//'
         wrap = lambda x: f'({x})' if isinstance(x, BinaryOp) else x
         return f'{wrap(self.expr)} {op} {wrap(self.separator)}'
 
@@ -123,6 +131,13 @@ class Alt(Expr):
 
         with pb.loop():
             with _if_fails(pb, self.expr):
+                # If we're not discarding separators, and if we're also not
+                # allowing a trailing separator, then we need to pop the last
+                # separator off of our list.
+                if not self.discard_separators and not self.allow_trailer:
+                    # But only pop if staging is not empty.
+                    with pb.IF(staging):
+                        pb(staging.pop())
                 pb(BREAK)
 
             pb(staging.append(RESULT))
@@ -131,6 +146,9 @@ class Alt(Expr):
             with _if_fails(pb, self.separator):
                 pb(BREAK)
 
+            if not self.discard_separators:
+                pb(staging.append(RESULT))
+
             if self.allow_trailer:
                 pb(checkpoint << POS)
 
@@ -234,12 +252,15 @@ class Choice(Expr):
     def _compile(self, pb):
         pb(Raw(f'# <{self.__class__.__name__}>'))
         backtrack = Var('backtrack')
-        farthest_pos = Var('farthest_pos')
 
-        pb(backtrack << farthest_pos << POS)
+        needs_err = not self.always_succeeds()
 
-        farthest_err = pb.var('farthest_err', Val(self.program_id))
-        farthest_err = pb.var('farthest_err', Raw(_error_func_name(self)))
+        if needs_err:
+            farthest_pos = Var('farthest_pos')
+            pb(backtrack << farthest_pos << POS)
+            farthest_err = pb.var('farthest_err', Raw(_error_func_name(self)))
+        else:
+            pb(backtrack << POS)
 
         with pb.breakable():
             for i, expr in enumerate(self.exprs):
@@ -248,15 +269,22 @@ class Choice(Expr):
                 with _if_succeeds(pb, expr):
                     pb(BREAK)
 
-                with pb.IF(farthest_pos < POS):
-                    pb(farthest_pos << POS)
-                    pb(farthest_err << RESULT)
+                if needs_err:
+                    if isinstance(expr, Fail):
+                        condition = farthest_pos <= POS
+                    else:
+                        condition = farthest_pos < POS
+
+                    with pb.IF(condition):
+                        pb(farthest_pos << POS)
+                        pb(farthest_err << RESULT)
 
                 if i + 1 < len(self.exprs):
                     pb(POS << backtrack)
 
-            pb(POS << farthest_pos)
-            pb(RESULT << farthest_err)
+            if needs_err:
+                pb(POS << farthest_pos)
+                pb(RESULT << farthest_err)
         pb(Raw(f'# </{self.__class__.__name__}>'))
 
     def complain(self):
@@ -413,6 +441,8 @@ class Fail(Expr):
     num_blocks = 0
 
     def __init__(self, message=None):
+        if isinstance(message, StringLiteral):
+            message = message.value
         self.message = message
 
     def __str__(self):
@@ -470,17 +500,29 @@ class LetExpression(Expr):
 class List(Expr):
     num_blocks = 2
 
-    def __init__(self, expr, allow_empty=True):
+    def __init__(self, expr, min_len=None, max_len=None):
         self.expr = expr
-        self.allow_empty = allow_empty
+        self.min_len = min_len
+        self.max_len = max_len
 
     def __str__(self):
-        x = f'({self.expr})' if isinstance(self.expr, BinaryOp) else self.expr
-        op = '*' if self.allow_empty else '+'
+        if isinstance(self.expr, (BinaryOp, LetExpression)):
+            x = f'({self.expr})'
+        else:
+            x = self.expr
+
+        if self.min_len is None and self.max_len is None:
+            op = '*'
+        elif self.min_len == 1 and self.max_len is None:
+            op = '+'
+        elif self.min_len == self.max_len:
+            op = f'{{{self.min_len}}}'
+        else:
+            op = f'{{{self.min_len},{self.max_len}}}'
         return f'{x}{op}'
 
     def always_succeeds(self):
-        return self.allow_empty
+        return not self.min_len or self.min_len == '0'
 
     def _compile(self, pb):
         pb(Raw(f'# <{self.__class__.__name__}>'))
@@ -495,15 +537,23 @@ class List(Expr):
 
             pb(staging.append(RESULT))
 
+            if self.max_len is not None:
+                with pb.IF(raw.len(staging) == Raw(self.max_len)):
+                    pb(BREAK)
+
         success = [
             RESULT << staging,
             STATUS << True,
         ]
 
-        if self.allow_empty:
+        if not self.min_len or self.min_len == '0':
             pb(*success)
         else:
-            with pb.IF(staging):
+            if self.min_len == 1 or self.min_len == '1':
+                condition = staging
+            else:
+                condition = raw.len(staging) >= Raw(self.min_len)
+            with pb.IF(condition):
                 pb(*success)
 
         pb(Raw(f'# </{self.__class__.__name__}>'))
@@ -563,16 +613,20 @@ class RegexLiteral(Expr):
     def __init__(self, pattern, ignore_case=False):
         if isinstance(pattern, typing.Pattern):
             pattern = pattern.pattern
-        if not isinstance(pattern, str):
-            raise TypeError('Expected str')
+        if not isinstance(pattern, (bytes, str)):
+            raise TypeError('Expected bytes or str')
         self.pattern = pattern
         self.skip_ignored = False
         self.ignore_case = ignore_case
 
     def __str__(self):
-        pattern = self.pattern.replace('\\', '\\\\')
+        pattern = self.pattern
+        if isinstance(pattern, bytes):
+            pattern = pattern.decode('ascii')
+
+        pattern = pattern.replace('\\', '\\\\')
         flag = 'i' if self.ignore_case else ''
-        return f'@/{pattern}/{flag}'
+        return f'/{pattern}/{flag}'
 
     def _compile(self, pb):
         pb(Raw(f'# <Regex pattern={self.pattern!r}>'))
@@ -724,26 +778,26 @@ class Skip(Expr):
 
 
 def Some(expr):
-    return List(expr, allow_empty=False)
+    return List(expr, min_len=1)
 
 
 class StringLiteral(Expr):
     def __init__(self, value):
-        if not isinstance(value, str):
-            raise TypeError(f'Expected str. Received: {type(value)}.')
+        if not isinstance(value, (bytes, str)):
+            raise TypeError(f'Expected bytes or str. Received: {type(value)}.')
         self.value = value
         self.skip_ignored = False
-        self.num_blocks = 0 if self.value == '' else 1
+        self.num_blocks = 0 if not self.value else 1
 
     def __str__(self):
         return repr(self.value)
 
     def always_succeeds(self):
-        return self.value == ''
+        return not self.value
 
     def _compile(self, pb):
         pb(Raw(f'# <String value={self.value!r}>'))
-        if self.value == '':
+        if not self.value:
             pb(STATUS << Val(True), RESULT << Val(''))
             pb(Raw('# </String>'))
             return
@@ -1016,7 +1070,7 @@ def _error_func_name(expr):
     return f'_raise_error{expr.program_id}'
 
 
-BinaryOp = (Alt, Apply, Choice, Discard, Where)
+BinaryOp = (Apply, Choice, Discard, Sep, Where)
 
 
 def generate_source_code(docstring, nodes):
@@ -1040,7 +1094,7 @@ def generate_source_code(docstring, nodes):
         if node.is_ignored:
             ignored.append(node)
 
-        if start_rule is None and node.name.lower() == 'start':
+        if start_rule is None and node.name and node.name.lower() == 'start':
             start_rule = node
 
     if start_rule is not None and start_rule.is_ignored:
@@ -1053,11 +1107,15 @@ def generate_source_code(docstring, nodes):
 
     visited_names = set()
     for rule in rules:
-        if rule.name.startswith('_'):
+        if rule.name is not None and rule.name.startswith('_'):
             raise Exception(
                 'Grammar rule names must start with a letter. Found a rule that'
                 f' starts with an underscore: "{rule.name}". '
             )
+
+        if not rule.name:
+            rule.name = f'_anonymous_{id(rule)}'
+
         if rule.name in visited_names:
             raise Exception(
                 'Each grammar rule must have a unique name. Found two or more'
@@ -1101,6 +1159,24 @@ def generate_source_code(docstring, nodes):
         start=_cont_name(default_rule.name),
     )))
 
+    error_delegates = {}
+    def set_error_delegate(expr):
+        if not isinstance(expr, Choice):
+            return
+        real, fail = [], []
+        for option in expr.exprs:
+            if isinstance(option, Fail):
+                fail.append(option)
+            else:
+                real.append(option)
+        if not real or not fail:
+            return
+        delegate = Choice(*real)
+        error_delegates[fail[-1].program_id] = Choice(*real)
+
+    for rule in rules:
+        visit(set_error_delegate, rule)
+
     visited = set()
     def maybe_compile_error_message(rule, expr):
         if not hasattr(expr, 'complain') or expr.program_id in visited:
@@ -1127,10 +1203,11 @@ def generate_source_code(docstring, nodes):
                     ),
                 )
 
+            delegate = error_delegates.get(expr.program_id, expr)
             pb(
                 raw.details << Raw('('),
                 Val(f'Failed to parse the {rule.name!r} rule, at the expression:\n'),
-                Val(f'    {str(expr)}\n\n'),
+                Val(f'    {str(delegate)}\n\n'),
                 Val(expr.complain()),
                 Raw(')'),
                 Raise(raw.ParseError(raw.title + raw.details, POS, raw.line, raw.col)),
@@ -1250,13 +1327,17 @@ class Rule:
 
 
 _main_template = r'''
-class ParseError(Exception):
+class SourcerError(Exception):
+    """Common superclass for ParseError and PartialParseError."""
+
+
+class ParseError(SourcerError):
     def __init__(self, message, index, line, column):
         super().__init__(message)
         self.position = _Position(index, line, column)
 
 
-class PartialParseError(Exception):
+class PartialParseError(SourcerError):
     def __init__(self, partial_result, last_position, excerpt):
         super().__init__('Incomplete parse. Unexpected input on line'
             f' {last_position.line}, column {last_position.column}:\n{excerpt}')
@@ -1423,6 +1504,9 @@ def _finalize_parse_info(text, nodes, pos, fullparse):
 
 
 def _extract_excerpt(text, pos, col):
+    if isinstance(text, bytes):
+        return repr(text[max(0, pos - 1) : pos + 2])
+
     start = pos - (col - 1)
     match = _compile_re('\n').search(text, pos + 1)
     end = len(text) if match is None else match.start()
